#!/usr/bin/perl

# PageRank calculator for directed graphs.
# Undirected graphs can also be handled; in this case please include all of the edges in both directions in the inputfile.
# Inputfile structure:
#	every line corresponds to an edge;
# 	every line is in the form of <source_nodeID> TAB <destination_nodeID> EOL;
#		this represents a directed edge pointing from <source_nodeID> to <destination_nodeID>
# 	Node IDs should not contain any whitespace characters.
# Integer edgeweights can be given by adding an edge several times.
# Damping factor has to be adjusted by modifying the program code: the $damping_factor variable serves this purpose.
#	It has to be set to a value between 0 and 1. Default: 0.15.
#	 When set to zero, the Markov-chain represented by the graph may not be irreducible (equivalently, the graph may not be strongly connected, meaning there may be two graph nodes that have no directed path between them)
#	For protein-protein networks with undirected edges, it may be advisable to set the damping factor to zero (no random restarting of the random walk on the graph). However, it can be proven that in this case pageranks will simply correlate with the degrees of nodes.
# Accuracy (and required runtime) can be changed by modifying the $epsilon variable.
#
# (c) Gabor IVAN, 2009

open (ATTR_DEGREE, "> OUT_PageRank__node_degrees.noa") or die "Could not open \"OUT_PageRank__node_degrees.noa\" for writing.\n\n";
open (ATTR_PAGERANK, "> OUT_PageRank__node_pageranks.noa") or die "Could not open \"OUT_PageRank__node_pageranks.noa\" for writing.\n\n";
open (INPUT_PERSONALIZATION, "< melanoma_markers_for_personalization.txt") or die "Could not open personalization file for reading.\n\n";

use IO::Handle;
STDOUT->autoflush(1);

$epsilon = 0.001;
# for the sake of simplicity (so that only nodes that have relevant personalization-weight have to be included in the inputfile), two damping factors are defined
# one ($damping_factor) is universal and the other is only relevant to nodes that have personalization info.
# if no nodes are given in the personalization input file, all nodes will have the same personalization weight,
# and the algorithm will work as if only one $damping_factor was present: the sum of $damping_factor and $personalization_damping_factor.
$damping_factor = 0.15;
$personalization_damping_factor = 0.15;

# Read the inputfile and construct the graph in memory.
%node_dstnodes = ();
%outdegree = ();
# ^^^a hash storing the outdegree of every point. As multiple edges between two nodes can be interpreted as one edge with a a weight of the number of this edge's occurrences,
# this equals to the sum of weight of edges pointing outwards from this particular node.  For a given NodeID, this further equals to the number this particular NodeID appears in the first column of the input file.
while (<>) {
	if (/(^\S+)\s(\S+)$/) { # this is a $1 ---> $2 edge
		if ($1 eq $2) {
			print "Skipping loop edge on '$1'.\n";
		} else {
			# count directed edges
			if (exists $outdegree{$1}) { $outdegree{$1}++; } else { $outdegree{$1} = 1; }
			if (not exists $outdegree{$2}) { $outdegree{$2} = 0; }
			# count "cumulative" edges (direction does not count)
			if (exists $degree{$1}) { $degree{$1}++; } else { $degree{$1} = 1; }
			if (exists $degree{$2}) { $degree{$2}++; } else { $degree{$2} = 1; }
			
			if (exists $node_dstnodes{$1}{$2}) {
				$node_dstnodes{$1}{$2}++;
				print "Warning: edge [$1 ---> $2] already exists. Weight increased to ".$node_dstnodes{$1}{$2}."\n";
			} else {
				$node_dstnodes{$1}{$2} = 1;
			}
		}
	}
}

# count the number of graph nodes
$N = 0; foreach (keys %degree) { $N++; }

# Read personalization file that contains weights for selected nodes of the graph.
$totalweight = 0;
while (<INPUT_PERSONALIZATION>) {
	if (/(^\S+)\s(\S+)/) {
		if (exists $degree{$1}) { # this is an existing node in the graph
			$node_personalizationweight{$1} = $2;
			$totalweight += $2;
		} else { print "Warning: during reading personalization data: node [$1] does not exist.\n"; }
	}
}
if ($totalweight > 0.0) { # if there is personalization data for at least one node...
	foreach (keys %degree) { # add a zero weight for personalization of nodes that do not exist in the personalization input file
		if (not exists $node_personalizationweight{$_}) { 
			$node_personalizationweight{$_} = 0.0;
		} else { $node_personalizationweight{$_} /= $totalweight; }
	}
} else { # set every node to weight 1/N.
	foreach (keys %degree) { $node_personalizationweight{$_} = 1.0 / $N; }
}

# Initialize variables: set initial PageRank value for each node to 1/N where N is the number of vertices.
%source_vector = (); # a note: the keys of %source_vector, %dest_vector and %outdegree will be the same
%dest_vector = ();
foreach (keys %outdegree) { $source_vector{$_} = 1.0 / $N; }
	print "\nPageRank algorithm initialized: $N nodes read from the input file.\nDamping factor for the algorithm is $damping_factor.\nOutdegree and personalization data for the nodes:\n\n";
	foreach (sort keys %outdegree) { print "\t$_\t".$outdegree{$_}."\t|\t".$node_personalizationweight{$_}."\n"; }
	print "\n\n";

# run the PageRank algorithm.
$residual = $epsilon + 1; # residual will be the squared norm of the vector %source_vector - %dest_vector
$iteration = 0;
while (sqrt $residual > $epsilon) {
	foreach (keys %outdegree) { $dest_vector{$_} = 0; }
	foreach $srcnode (keys %node_dstnodes) { # we go through every node that has outgoing edges.
		foreach $dstnode (keys %{$node_dstnodes{$srcnode}}) { # we give a vote to every $dstnode that is pointed by $srcnode;
			$dest_vector{$dstnode} += $source_vector{$srcnode} * $node_dstnodes{$srcnode}{$dstnode} / $outdegree{$srcnode};
			# ^^^ give a vote to $dstnode weighted by the edgeweight of the $srcnode ---> $dstnode edge and normalized by the sum of weights (stored in $outdegree)
		}
	}
	foreach (keys %dest_vector) { # correction using the $damping_factor and personalization data
		$dest_vector{$_} = (1 - $damping_factor - $personalization_damping_factor) * $dest_vector{$_} + $damping_factor / $N + $personalization_damping_factor * $node_personalizationweight{$_};
	}
	$residual = 0;
	foreach (keys %source_vector) { $residual += ($source_vector{$_} - $dest_vector{$_}) * ($source_vector{$_} - $dest_vector{$_}); }
	foreach (keys %source_vector) { $source_vector{$_} = $dest_vector{$_}; }
	$iteration++;
	printf ("PageRank: $iteration. iteration, residual is currently %10.8f, (target is $epsilon).\n", sqrt $residual);
}

# output the graph nodes and the PageRank values assigned to them.
#print "\n\nResults of PageRank calculation [NodeID - assigned PageRank value], NodeIDs are alphabetically sorted:\n\n";
#foreach (sort keys %source_vector) { print "\t$_\t".(100 * $source_vector{$_})." %\n"; }
#print "\n\nResults of PageRank calculation [PageRank value - NodeID], PageRanks are in descending order:\n\n";
%tmp_vector = ();
$min_pagerank = 1.0;
foreach (sort keys %source_vector) {
	$tmp = sprintf ("%10.9f|%s", $source_vector{$_}, $_);
	$tmp_vector{$tmp} = 1;
	if ($source_vector{$_} < $min_pagerank) { $min_pagerank = $source_vector{$_}; }
}
print ATTR_DEGREE "Degree\n";
print ATTR_PAGERANK "PageRank (class=Double)\n";
foreach (reverse sort keys %tmp_vector) {
	if (/^([^|]+)\|([^|]+)$/) {
		#print "\t$1\t$2\n";
		printf ATTR_DEGREE ("$2 = %d\n", $degree{$2});
		printf ATTR_PAGERANK ("$2 = %12.2f\n", $1 / $min_pagerank);
	}
}
print "\n\nResults of PageRank calculation [Normalized PageRank value - NodeID],\n";
printf ( "sum of PageRanks is now %12.1f (instead of 1.0), PageRanks are in descending order:\n\n", 1.0 / $min_pagerank);
print "   \t    PageRank     |      NodeID      |     Degree    | Additional weight from damping | Additional weight for personalization\n";
printf ("   \t                 |                  |               |      (weighted by: %4.3f)      |   (weighted by: %4.3f)\n===============================================================================================================================\n\n", $damping_factor, $personalization_damping_factor);
foreach (reverse sort keys %tmp_vector) {
	if (/^([^|]+)\|([^|]+)$/) {
		my $uID = $2;
		printf ("-->\t%12.2f     |     %8s     |     ", $1 / $min_pagerank, $uID);
		#if (exists $uniprotid_species{$uID}) { printf ("%6d     |    ", $uniprotid_species{$uID}); } else { print "  -        |    "; }
		printf ("%6d    |            ", $outdegree{$uID});
		printf ("%8.7f           |    ", 1 / $N);
		if (exists $node_personalizationweight{$uID}) { print $node_personalizationweight{$uID}."\n"; } else { print "-\n"; }
	}
}


# filter multiple occurring scalars from a scalar-list. Return the filtered list.
sub filter_list () {

	my @output;
	my %tmp_hash = ();
	foreach (@_) { $tmp_hash{$_} = 1; }
	foreach (sort keys %tmp_hash) { push (@output, $_); }
	return @output;

}
